/*
 * Copyright (c) 2018-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * See file LICENSE for terms.
 */

#ifndef PARALLEL_TREE_FINDER_H
#define PARALLEL_TREE_FINDER_H

#include "thread_pool.h"
#include "agg_tree.h"
#include "dump_file.h"

class ParallelTreeFinder;

using ListTreeFinderTasksPtr = std::list<class TreeFinderTask*>;

// A single task of tree finding, checking whether a single tree can match
class TreeFinderTask : public ThreadPoolTask {

    AggTree                &m_agg_tree_;
    ParallelTreeFinder     &m_tree_finder_;

    JobSubTreeInfo m_job_sub_tree_info_;
    JobSubTreeScore m_sub_tree_result_;

public:
    TreeFinderTask(AggTree &agg_tree, ParallelTreeFinder &tree_finder) :
        m_agg_tree_(agg_tree), m_tree_finder_(tree_finder) {}

    virtual ~TreeFinderTask() {}

    JobSubTreeInfo &GetJobSubTreeInfo() { return m_job_sub_tree_info_; }
    JobSubTreeScore &GetJobSubTreeScore() { return m_sub_tree_result_; }
    
    sharp_trees_t GetTreeId() const { return m_agg_tree_.GetId(); }

    virtual void Run();
};

//Invoke method on all trees In parallel using multi threading
class ParallelTreeFinder : public ThreadPoolTasksCollection {

    ThreadPool              m_thread_pool_;
    ListTreeFinderTasksPtr  m_tasks_;

    //search data
    const SetPortDataConstPtr  *m_compute_ports_;
    const JobResource          *m_job_resource_;
    
    std::unique_ptr<file_utils::DumpFile> m_job_failures_dump_file_ptr_;

public:

    ParallelTreeFinder() : ThreadPoolTasksCollection(),
        m_thread_pool_(),
        m_compute_ports_(NULL), 
        m_job_resource_(NULL),
        m_job_failures_dump_file_ptr_{file_utils::GetDumpFileIfEnabled("sharp_am_failed_job_requests_details.dump")} {}

    ~ParallelTreeFinder();

    void Init();

    void GetJobSubTreeInfo(const SetPortDataConstPtr& compute_ports,
                           const JobResource &job_resource,
                           const SharpJob &p_job,
                           const sharp_job_id_t job_id,
                           JobSubTreeInfo &best_tree_info,
                           JobSubTreeScore &best_tree_result);

    void DumpJobRequestFailure(const uint64_t job_id,
                               JobSubTreeInfo const * const best_tree_info = nullptr,
                               char const * const failure_reason = nullptr);

    const SetPortDataConstPtr *GetComputePorts() { return m_compute_ports_; }
    const JobResource *GetJobResource() {return m_job_resource_; }

    int CreateLltTreeForJob(const SharpJob &p_job,
                            JobSubTreeInfo &job_sub_tree_info);

    int CreateSatTreeForJob(const SharpJob &p_job,
                            const JobResource &job_resource,
                            JobSubTreeInfo &job_sub_tree_info,
                            JobSubTreeScore &sub_tree_result);

    void ReCreateTasks();
};

#endif //PARALLEL_TREE_FINDER_H
